;; Generated package description from mode-line-maker.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "mode-line-maker" "0.1.0.20251206.64613" "Mode-line helper tools" '((emacs "29.1")) :commit "8d70286c3fda845f3b8d97a937248670ca002f13" :maintainer '("Nicolas P. Rougier" . "Nicolas.Rougier@inria.fr") :keywords '("convenience" "mode-line" "header-line") :url "https://github.com/rougier/mode-line-maker")
